require({cache:{
'url:sas-cpb/datagrid/ActionableGridTemplate.html':"<div>\n   <div id='gridwrap_${_gridArgs.cpb_name}' class='mainTable'>\n      <div data-dojo-attach-point='gridTitle' class='dataGridTitle'></div>\n      <div id='gridAndToolbarDiv_${_gridArgs.cpb_name}'>\n         <span id='gridToolbar_${_gridArgs.cpb_name}'></span>\n         <div id='${_gridArgs.cpb_name}' data-dojo-attach-point='gridDiv'></div>\n      </div>\n   </div>\n</div>"}});
/**
 * A composite component consisting of a grid and (optional) grid-level and row-level actions.
 * Grid-level actions are displayed in a tool bar above the grid.  Row-level actions are
 * displayed column on the left-hand side of the grid.  If any grid-level actions require rows
 * to be selected, a checkbox column is automatically added and linked to those actions.
 * An optional title can also be specified.
 */
define("sas-cpb/datagrid/ActionableGrid", ['dojo/_base/declare', 'dojo/_base/array',
      'dojo/dom-style',
      'dijit/Toolbar',
      'dgrid/selector',
      'dijit/_WidgetBase',
      'dijit/_TemplatedMixin',
      'dijit/_WidgetsInTemplateMixin',
      'dijit/layout/_LayoutWidget',
      'sas-cpb/datagrid/_BaseComponentGrid',
      'sas-cpb/datagrid/CPBDataGrid', 'sas-cpb/datagrid/CPBGridUtil',
      'dojo/text!./ActionableGridTemplate.html'],
   function(declare, array,
         domStyle,
         Toolbar,
         selector,
         _WidgetBase, _TemplatedMixin, _WidgetsInTemplateMixin, _LayoutWidget,
         _BaseComponentGrid,
         CPBDataGrid, CPBGridUtil,
         template) {

var getSelectorType = function(args, gridActions) {
   if (args.selectorType && args.selectorType!=='default') {
      return args.selectorType;
   }
   if (gridActions && gridActions.length>0) {
      for (var i=0; i<gridActions.length; i++) {
         if (gridActions[i].sas_selectionButton) {
            return 'checkbox';
         }
      }
   }
   return 'none';
};

var isEmpty = function(obj) {
   for (var prop in obj) {
      if(obj.hasOwnProperty(prop))
         return false;
   }
   return true;
};

var ActionableGrid = declare('ActionableGrid', [_WidgetBase, _LayoutWidget, _TemplatedMixin, _WidgetsInTemplateMixin], {

   templateString: template,
   widgetsInTemplate: true,
   grid: null,
   _title: null,
   baseClass: 'cpbGrid',
   
   constructor: function(args) {
      this._prepareGridArguments(args);
   },
   
   layout: function() {
      this.grid.resize(); // So grid in e.g. liveSplitter'd BorderContainer resizes properly
   },

   buildRendering: function() {
      this.inherited(arguments);
   },
   
   postCreate: function() {
      
      this.inherited(arguments);
      
      // Create our grid
      this.grid = this._createGrid();
      delete this._gridArgs;
      
      // Create tool bar containing grid-level actions
      if (this.gridActions && this.gridActions.length>0) {
         var toolbarId = "gridToolbar_" + this.grid.cpb_name;
         var toolbar = new Toolbar({}, toolbarId);
         for (var i = 0; i < this.gridActions.length; i++) {
            toolbar.addChild(this.gridActions[i]);
         }
         
         var selectionChangedHandler = function(event) {
            // Update any grid actions dependent on a selection
            var selection = event.grid.selection;
            var toolbarId = "gridToolbar_" + event.grid.cpb_name;
            var toolbar = dijit.byId(toolbarId);
            if (toolbar) {
               var selection = !isEmpty(selection);
               var children = toolbar.getChildren();
               for (var i = 0; i < children.length; i++) {
                  if (children[i].sas_selectionButton) {
                     children[i].set('disabled', !selection);
                  }
               }
            }
         };
         this.grid.on('dgrid-select', selectionChangedHandler);
         this.grid.on('dgrid-deselect', selectionChangedHandler);
       
      }
      
      this.set('title', this._title); // Hack to get it to actually display
      
   },
   
   startup: function() {
      this.inherited(arguments);
      // When loaded in a lazy tab, clicking the select all checkbox throws a NPE. This is because
      // grid._rowIdToObject returns a row for the header row of the grid when it shouldn't, which prevents the
      // appropriate refresh method that reset grid._rowIdToObject to an empty object from executing. The workaround
      // is to reset this grid._rowIdToObject to {} before calling the grid's startup method. This is a crucial step
      // in rendering the graph that is omitted by failing to call refresh() from dgrid's List.
      this.grid._rowIdToObject = {};
      this.grid.startup();
   },
   
   _createGrid: function() {
      return new _BaseComponentGrid(this._gridArgs, this.gridDiv);
   },
   
   _prepareGridArguments: function(args) {
      
      var gridModel = args.gridModel;
      
      // Create grid actions
      var formattedGridName = CPBGridUtil.getFormattedUINameFromID(gridModel.name);
      var gridActionsModel = this.createGridActions(formattedGridName, gridModel);
      var extraGridActions = args.actions;
      if (extraGridActions) {
         if (gridActionsModel) {
            gridActionsModel = gridActionsModel.concat(extraGridActions);
         }
         else {
            gridActionsModel = extraGridActions;
         }
      }
      var result = CPBGridUtil.createGridActions(gridModel.name, gridActionsModel);
      this.gridActions = result.gridActions;
      var rowActions = result.rowActions;
      var refreshListeners = result.refreshListeners;
      
      // Determine whether we need a selection column
      var selectorColumn = null,
          selectionMode = null;
      if (!gridModel.readOnly) {
         var selectorType = getSelectorType(gridModel, this.gridActions);
         if (selectorType!=='none') {
            var columnParams = { field: 'checkboxes', selectorType: selectorType, resizable: false, unhidable: true, sortable: false, reorderable: false };
            if (selectorType==='checkbox') {
               selectionMode = 'multiple';
            }
            else {
               selectionMode = 'single';
               columnParams.label = ' '; // Workaround for dgrid bug
            }
            selectorColumn = selector(columnParams);
         }
      }
      if (!selectionMode) {
         selectionMode = 'single';
      }
      
      // Create grid columns
      var columnModel = args.columns;
      if (selectorColumn) {
         columnModel.splice(0, 0, selectorColumn);
      }
      var columns = CPBGridUtil.createGridColumns(formattedGridName, columnModel,
            rowActions, gridModel.lazyRowActionsUrl, selectorColumn!=null);
      
      // Create the store
      var store;
      if (args.storeData) {
         // Sometimes we just need all the data to be local immediately, such as print mode (don't want saved
         // HTML to hit the server).
         //var json = JSON.parse(jsonString, false);
         store = CPBDataGrid.createMemoryStore(formattedGridName, args.storeData);//json);
      }
      else {
         var cacheDataInUIContext = gridModel.cacheDataInUIContext;
         var url = CPBGridUtil.getURL(formattedGridName, gridModel.uiContextId, cacheDataInUIContext);
         store = CPBDataGrid.createStore(formattedGridName, url, cacheDataInUIContext);
      }
      
      var _addAllPageSize = false;
      if (gridModel.pageSizeOptions && gridModel.pageSizeOptions.length) {
         
         // dgrid has no means to provide labels for its select options, so we take special care for 'all'
         var index = array.indexOf(gridModel.pageSizeOptions, 'all');
         if (index > -1) {
            gridModel.pageSizeOptions.splice(index, 1);
            _addAllPageSize = true;
         }
         
         // Make sure the initial page size is in the size choice list
         if (gridModel.pageSize && array.indexOf(gridModel.pageSizeOptions, gridModel.pageSize)==-1) {
            gridModel.pageSizeOptions.push(gridModel.pageSize);
         }
         gridModel.pageSizeOptions.sort(function(a, b) { return a<b ? -1 : (a>b ? 1 : 0) });
         
      }
      
      this._gridArgs = {
         id: gridModel.name,
         columns: columns,
         store: store,
         initiallyLoadData: (args.initiallyLoadData === undefined) ? true : args.initiallyLoadData,
         selection: {},
         selectionMode: selectionMode,
         allowSelectAll: true,
         deselectOnRefresh: true,
         loadingMessage: '<div style="display: inline-block" class="dijitIconLoading"></div> ' + AppProperties['table.default.loading.message.txt'],
         noDataMessage: AppProperties['table.default.no.rows.message.txt'],
         visible: gridModel.visible,
         cpb_id: gridModel.id,
         cpb_name: formattedGridName,
         cpb_storeName: gridModel.storeName,
         cpb_hasOnChangeEvents: gridModel.fireOnChangeEvents,
         fireOnSelectEvents: gridModel.hasOnSelectActions,//true,
         cpb_selectedKeyField: gridModel.selectedKeyField,
         rowsPerPage: gridModel.printMode ? 100000000 : gridModel.pageSize,
         firstLastArrows: gridModel.printMode ? false : true,
         pageSizeOptions: gridModel.printMode ? [] : gridModel.pageSizeOptions || [],
         _addAllPageSize: _addAllPageSize,
         pagingLinks: gridModel.printMode ? 0 : 2,
         printMode: gridModel.printMode || false,
         refreshListeners: refreshListeners
      };
      
      // Style CSS to set on the child grid
      if (args.gridStyle) {
         this._gridArgs.style = args.gridStyle;
      }
      
      if (args.title) {
         this.set('title', args.title);
      }
      
   },
   
   // summary:
   //    Getter for the actual grid component in this composite component.
   _getGridAttr: function() {
      return this.grid;
   },
   
   _getTitleAttr: function() {
      return this._title;
   },
   _setTitleAttr: function(title) {
      this._title = title;
      if (this.gridTitle) {
         this.gridTitle.innerHTML = '';
         if (title) {
            this.gridTitle.appendChild(document.createTextNode(title));
            domStyle.set(this.gridTitle, 'display', 'block');
         }
         else {
            domStyle.set(this.gridTitle, 'display', 'none');
         }
      }
   },
   
   // summary:
   //    Subclasses can override this method to return actions for the tool bar.
   //    If any grid actions were specified as parameters, they will be appended to
   //    this list.
   createGridActions: function(gridName) {
      return null;
   }
   
});

return ActionableGrid;

});
